@@ -17,6 +17,7 @@
 //!
 //! Features denoted as experimental in the C library are hidden behind an
 //! `experimental` feature.
+#![cfg_attr(feature = "doc-cfg", feature(doc_cfg))]
 
 #[cfg(feature = "std")]
 extern crate std;
@@ -40,8 +41,8 @@ use std::os::raw::{c_char, c_int, c_ulonglong, c_void};
 use libc::{c_char, c_int, c_ulonglong, c_void};
 
 use core::marker::PhantomData;
-use core::ops::Deref;
-use core::ops::DerefMut;
+use core::ops::{Deref, DerefMut};
+use core::ptr::NonNull;
 use core::str;
 
 include!("constants.rs");
@@ -152,18 +153,33 @@ pub fn compress_bound(src_size: usize) -> usize {
     unsafe { zstd_sys::ZSTD_compressBound(src_size) }
 }
 
-pub struct CCtx<'a>(*mut zstd_sys::ZSTD_CCtx, PhantomData<&'a ()>);
+pub struct CCtx<'a>(NonNull<zstd_sys::ZSTD_CCtx>, PhantomData<&'a ()>);
 
-impl<'a> Default for CCtx<'a> {
+impl Default for CCtx<'_> {
     fn default() -> Self {
-        create_cctx()
+        CCtx::create()
     }
 }
 
 impl CCtx<'static> {
+    /// Tries to create a new context.
+    ///
+    /// Returns `None` if zstd returns a NULL pointer - may happen if allocation fails.
+    pub fn try_create() -> Option<Self> {
+        Some(CCtx(
+            NonNull::new(unsafe { zstd_sys::ZSTD_createCCtx() })?,
+            PhantomData,
+        ))
+    }
+
     /// Wrap `ZSTD_createCCtx`
+    ///
+    /// # Panics
+    ///
+    /// If zstd returns a NULL pointer.
     pub fn create() -> Self {
-        CCtx(unsafe { zstd_sys::ZSTD_createCCtx() }, PhantomData)
+        Self::try_create()
+            .expect("zstd returned null pointer when creating new context")
     }
 }
 
@@ -178,7 +194,7 @@ impl<'a> CCtx<'a> {
         unsafe {
             dst.write_from(|buffer, capacity| {
                 parse_code(zstd_sys::ZSTD_compressCCtx(
-                    self.0,
+                    self.0.as_ptr(),
                     buffer,
                     capacity,
                     ptr_void(src),
@@ -198,7 +214,7 @@ impl<'a> CCtx<'a> {
         unsafe {
             dst.write_from(|buffer, capacity| {
                 parse_code(zstd_sys::ZSTD_compress2(
-                    self.0,
+                    self.0.as_ptr(),
                     buffer,
                     capacity,
                     ptr_void(src),
@@ -219,7 +235,7 @@ impl<'a> CCtx<'a> {
         unsafe {
             dst.write_from(|buffer, capacity| {
                 parse_code(zstd_sys::ZSTD_compress_usingDict(
-                    self.0,
+                    self.0.as_ptr(),
                     buffer,
                     capacity,
                     ptr_void(src),
@@ -242,23 +258,26 @@ impl<'a> CCtx<'a> {
         unsafe {
             dst.write_from(|buffer, capacity| {
                 parse_code(zstd_sys::ZSTD_compress_usingCDict(
-                    self.0,
+                    self.0.as_ptr(),
                     buffer,
                     capacity,
                     ptr_void(src),
                     src.len(),
-                    cdict.0,
+                    cdict.0.as_ptr(),
                 ))
             })
         }
     }
 
     pub fn init(&mut self, compression_level: CompressionLevel) -> usize {
-        unsafe { zstd_sys::ZSTD_initCStream(self.0, compression_level) }
+        unsafe {
+            zstd_sys::ZSTD_initCStream(self.0.as_ptr(), compression_level)
+        }
     }
 
     /// Wraps the `ZSTD_initCStream_srcSize()` function.
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     #[deprecated]
     pub fn init_src_size(
         &mut self,
@@ -267,7 +286,7 @@ impl<'a> CCtx<'a> {
     ) -> usize {
         unsafe {
             zstd_sys::ZSTD_initCStream_srcSize(
-                self.0,
+                self.0.as_ptr(),
                 compression_level as c_int,
                 pledged_src_size as c_ulonglong,
             )
@@ -276,6 +295,7 @@ impl<'a> CCtx<'a> {
 
     /// Wraps the `ZSTD_initCStream_usingDict()` function.
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     #[deprecated]
     pub fn init_using_dict(
         &mut self,
@@ -284,7 +304,7 @@ impl<'a> CCtx<'a> {
     ) -> SafeResult {
         let code = unsafe {
             zstd_sys::ZSTD_initCStream_usingDict(
-                self.0,
+                self.0.as_ptr(),
                 ptr_void(dict),
                 dict.len(),
                 compression_level,
@@ -295,20 +315,25 @@ impl<'a> CCtx<'a> {
 
     /// Wraps the `ZSTD_initCStream_usingCDict()` function.
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     #[deprecated]
     pub fn init_using_cdict<'b>(&mut self, cdict: &CDict<'b>) -> SafeResult
     where
         'b: 'a, // Dictionary outlives the stream.
     {
-        let code =
-            unsafe { zstd_sys::ZSTD_initCStream_usingCDict(self.0, cdict.0) };
+        let code = unsafe {
+            zstd_sys::ZSTD_initCStream_usingCDict(
+                self.0.as_ptr(),
+                cdict.0.as_ptr(),
+            )
+        };
         parse_code(code)
     }
 
     pub fn load_dictionary(&mut self, dict: &[u8]) -> SafeResult {
         parse_code(unsafe {
             zstd_sys::ZSTD_CCtx_loadDictionary(
-                self.0,
+                self.0.as_ptr(),
                 ptr_void(dict),
                 dict.len(),
             )
@@ -322,7 +347,9 @@ impl<'a> CCtx<'a> {
     where
         'b: 'a,
     {
-        parse_code(unsafe { zstd_sys::ZSTD_CCtx_refCDict(self.0, cdict.0) })
+        parse_code(unsafe {
+            zstd_sys::ZSTD_CCtx_refCDict(self.0.as_ptr(), cdict.0.as_ptr())
+        })
     }
 
     pub fn ref_prefix<'b>(&mut self, prefix: &'b [u8]) -> SafeResult
@@ -331,7 +358,7 @@ impl<'a> CCtx<'a> {
     {
         parse_code(unsafe {
             zstd_sys::ZSTD_CCtx_refPrefix(
-                self.0,
+                self.0.as_ptr(),
                 ptr_void(prefix),
                 prefix.len(),
             )
@@ -347,7 +374,7 @@ impl<'a> CCtx<'a> {
         let mut input = input.wrap();
         let code = unsafe {
             zstd_sys::ZSTD_compressStream(
-                self.0,
+                self.0.as_ptr(),
                 ptr_mut(&mut output),
                 ptr_mut(&mut input),
             )
@@ -366,7 +393,7 @@ impl<'a> CCtx<'a> {
         let mut input = input.wrap();
         parse_code(unsafe {
             zstd_sys::ZSTD_compressStream2(
-                self.0,
+                self.0.as_ptr(),
                 ptr_mut(&mut output),
                 ptr_mut(&mut input),
                 end_op,
@@ -381,7 +408,7 @@ impl<'a> CCtx<'a> {
     ) -> SafeResult {
         let mut output = output.wrap();
         let code = unsafe {
-            zstd_sys::ZSTD_flushStream(self.0, ptr_mut(&mut output))
+            zstd_sys::ZSTD_flushStream(self.0.as_ptr(), ptr_mut(&mut output))
         };
         parse_code(code)
     }
@@ -392,25 +419,29 @@ impl<'a> CCtx<'a> {
         output: &mut OutBuffer<'_, C>,
     ) -> SafeResult {
         let mut output = output.wrap();
-        let code =
-            unsafe { zstd_sys::ZSTD_endStream(self.0, ptr_mut(&mut output)) };
+        let code = unsafe {
+            zstd_sys::ZSTD_endStream(self.0.as_ptr(), ptr_mut(&mut output))
+        };
         parse_code(code)
     }
 
     pub fn sizeof(&self) -> usize {
-        unsafe { zstd_sys::ZSTD_sizeof_CCtx(self.0) }
+        unsafe { zstd_sys::ZSTD_sizeof_CCtx(self.0.as_ptr()) }
     }
 
     pub fn reset(&mut self, reset: ResetDirective) -> SafeResult {
-        parse_code(unsafe { zstd_sys::ZSTD_CCtx_reset(self.0, reset) })
+        parse_code(unsafe {
+            zstd_sys::ZSTD_CCtx_reset(self.0.as_ptr(), reset)
+        })
     }
 
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     #[deprecated]
     pub fn reset_cstream(&mut self, pledged_src_size: u64) -> SafeResult {
         let code = unsafe {
             zstd_sys::ZSTD_resetCStream(
-                self.0,
+                self.0.as_ptr(),
                 pledged_src_size as c_ulonglong,
             )
         };
@@ -518,7 +549,7 @@ impl<'a> CCtx<'a> {
         };
 
         parse_code(unsafe {
-            zstd_sys::ZSTD_CCtx_setParameter(self.0, param, value)
+            zstd_sys::ZSTD_CCtx_setParameter(self.0.as_ptr(), param, value)
         })
     }
 
@@ -528,20 +559,46 @@ impl<'a> CCtx<'a> {
     ) -> SafeResult {
         parse_code(unsafe {
             zstd_sys::ZSTD_CCtx_setPledgedSrcSize(
-                self.0,
+                self.0.as_ptr(),
                 pledged_src_size as c_ulonglong,
             )
         })
     }
 
+    /// Creates a copy of this context.
+    ///
+    /// This only works before any data has been compressed. An error will be
+    /// returned otherwise.
+    #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
+    pub fn try_clone(
+        &self,
+        pledged_src_size: Option<u64>,
+    ) -> Result<Self, ErrorCode> {
+        let context = NonNull::new(unsafe { zstd_sys::ZSTD_createCCtx() })
+            .ok_or(0usize)?;
+
+        parse_code(unsafe {
+            zstd_sys::ZSTD_copyCCtx(
+                context.as_ptr(),
+                self.0.as_ptr(),
+                pledged_src_size.unwrap_or(CONTENTSIZE_UNKNOWN),
+            )
+        })?;
+
+        Ok(CCtx(context, self.1))
+    }
+
     /// Wraps the `ZSTD_getBlockSize()` function.
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     pub fn get_block_size(&self) -> usize {
-        unsafe { zstd_sys::ZSTD_getBlockSize(self.0) }
+        unsafe { zstd_sys::ZSTD_getBlockSize(self.0.as_ptr()) }
     }
 
     /// Wraps the `ZSTD_compressBlock()` function.
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     pub fn compress_block<C: WriteBuf + ?Sized>(
         &mut self,
         dst: &mut C,
@@ -550,7 +607,7 @@ impl<'a> CCtx<'a> {
         unsafe {
             dst.write_from(|buffer, capacity| {
                 parse_code(zstd_sys::ZSTD_compressBlock(
-                    self.0,
+                    self.0.as_ptr(),
                     buffer,
                     capacity,
                     ptr_void(src),
@@ -575,7 +632,7 @@ pub fn create_cctx<'a>() -> CCtx<'a> {
 impl<'a> Drop for CCtx<'a> {
     fn drop(&mut self) {
         unsafe {
-            zstd_sys::ZSTD_freeCCtx(self.0);
+            zstd_sys::ZSTD_freeCCtx(self.0.as_ptr());
         }
     }
 }
@@ -636,17 +693,24 @@ pub fn compress2(
 /// If no dictionary was used, it will most likely be `'static`.
 ///
 /// Same as `DStream`.
-pub struct DCtx<'a>(*mut zstd_sys::ZSTD_DCtx, PhantomData<&'a ()>);
+pub struct DCtx<'a>(NonNull<zstd_sys::ZSTD_DCtx>, PhantomData<&'a ()>);
 
 impl Default for DCtx<'_> {
     fn default() -> Self {
-        create_dctx()
+        DCtx::create()
     }
 }
 
 impl DCtx<'static> {
+    pub fn try_create() -> Option<Self> {
+        Some(DCtx(
+            NonNull::new(unsafe { zstd_sys::ZSTD_createDCtx() })?,
+            PhantomData,
+        ))
+    }
     pub fn create() -> Self {
-        DCtx(unsafe { zstd_sys::ZSTD_createDCtx() }, PhantomData)
+        Self::try_create()
+            .expect("zstd returned null pointer when creating new context")
     }
 }
 
@@ -660,7 +724,7 @@ impl<'a> DCtx<'a> {
         unsafe {
             dst.write_from(|buffer, capacity| {
                 parse_code(zstd_sys::ZSTD_decompressDCtx(
-                    self.0,
+                    self.0.as_ptr(),
                     buffer,
                     capacity,
                     ptr_void(src),
@@ -680,7 +744,7 @@ impl<'a> DCtx<'a> {
         unsafe {
             dst.write_from(|buffer, capacity| {
                 parse_code(zstd_sys::ZSTD_decompress_usingDict(
-                    self.0,
+                    self.0.as_ptr(),
                     buffer,
                     capacity,
                     ptr_void(src),
@@ -702,12 +766,12 @@ impl<'a> DCtx<'a> {
         unsafe {
             dst.write_from(|buffer, capacity| {
                 parse_code(zstd_sys::ZSTD_decompress_usingDDict(
-                    self.0,
+                    self.0.as_ptr(),
                     buffer,
                     capacity,
                     ptr_void(src),
                     src.len(),
-                    ddict.0,
+                    ddict.0.as_ptr(),
                 ))
             })
         }
@@ -717,16 +781,17 @@ impl<'a> DCtx<'a> {
     ///
     /// Initializes an existing `DStream` for decompression.
     pub fn init(&mut self) -> usize {
-        unsafe { zstd_sys::ZSTD_initDStream(self.0) }
+        unsafe { zstd_sys::ZSTD_initDStream(self.0.as_ptr()) }
     }
 
     /// Wraps the `ZSTD_initDStream_usingDict()` function.
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     #[deprecated]
     pub fn init_using_dict(&mut self, dict: &[u8]) -> SafeResult {
         let code = unsafe {
             zstd_sys::ZSTD_initDStream_usingDict(
-                self.0,
+                self.0.as_ptr(),
                 ptr_void(dict),
                 dict.len(),
             )
@@ -736,13 +801,18 @@ impl<'a> DCtx<'a> {
 
     /// Wraps the `ZSTD_initDStream_usingDDict()` function.
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     #[deprecated]
     pub fn init_using_ddict<'b>(&mut self, ddict: &DDict<'b>) -> SafeResult
     where
         'b: 'a,
     {
-        let code =
-            unsafe { zstd_sys::ZSTD_initDStream_usingDDict(self.0, ddict.0) };
+        let code = unsafe {
+            zstd_sys::ZSTD_initDStream_usingDDict(
+                self.0.as_ptr(),
+                ddict.0.as_ptr(),
+            )
+        };
         parse_code(code)
     }
 
@@ -750,7 +820,7 @@ impl<'a> DCtx<'a> {
     pub fn reset(&mut self) -> SafeResult {
         let code = unsafe {
             zstd_sys::ZSTD_DCtx_reset(
-                self.0,
+                self.0.as_ptr(),
                 ResetDirective::ZSTD_reset_session_only,
             )
         };
@@ -760,7 +830,7 @@ impl<'a> DCtx<'a> {
     pub fn load_dictionary(&mut self, dict: &[u8]) -> SafeResult {
         parse_code(unsafe {
             zstd_sys::ZSTD_DCtx_loadDictionary(
-                self.0,
+                self.0.as_ptr(),
                 ptr_void(dict),
                 dict.len(),
             )
@@ -771,7 +841,9 @@ impl<'a> DCtx<'a> {
     where
         'b: 'a,
     {
-        parse_code(unsafe { zstd_sys::ZSTD_DCtx_refDDict(self.0, ddict.0) })
+        parse_code(unsafe {
+            zstd_sys::ZSTD_DCtx_refDDict(self.0.as_ptr(), ddict.0.as_ptr())
+        })
     }
 
     pub fn ref_prefix<'b>(&mut self, prefix: &'b [u8]) -> SafeResult
@@ -780,7 +852,7 @@ impl<'a> DCtx<'a> {
     {
         parse_code(unsafe {
             zstd_sys::ZSTD_DCtx_refPrefix(
-                self.0,
+                self.0.as_ptr(),
                 ptr_void(prefix),
                 prefix.len(),
             )
@@ -819,7 +891,7 @@ impl<'a> DCtx<'a> {
         };
 
         parse_code(unsafe {
-            zstd_sys::ZSTD_DCtx_setParameter(self.0, param, value)
+            zstd_sys::ZSTD_DCtx_setParameter(self.0.as_ptr(), param, value)
         })
     }
 
@@ -833,7 +905,7 @@ impl<'a> DCtx<'a> {
         let mut input = input.wrap();
         let code = unsafe {
             zstd_sys::ZSTD_decompressStream(
-                self.0,
+                self.0.as_ptr(),
                 ptr_mut(&mut output),
                 ptr_mut(&mut input),
             )
@@ -857,11 +929,12 @@ impl<'a> DCtx<'a> {
 
     /// Wraps the `ZSTD_sizeof_DCtx()` function.
     pub fn sizeof(&self) -> usize {
-        unsafe { zstd_sys::ZSTD_sizeof_DCtx(self.0) }
+        unsafe { zstd_sys::ZSTD_sizeof_DCtx(self.0.as_ptr()) }
     }
 
     /// Wraps the `ZSTD_decompressBlock()` function.
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     pub fn decompress_block<C: WriteBuf + ?Sized>(
         &mut self,
         dst: &mut C,
@@ -870,7 +943,7 @@ impl<'a> DCtx<'a> {
         unsafe {
             dst.write_from(|buffer, capacity| {
                 parse_code(zstd_sys::ZSTD_decompressBlock(
-                    self.0,
+                    self.0.as_ptr(),
                     buffer,
                     capacity,
                     ptr_void(src),
@@ -882,11 +955,31 @@ impl<'a> DCtx<'a> {
 
     /// Wraps the `ZSTD_insertBlock()` function.
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     pub fn insert_block(&mut self, block: &[u8]) -> usize {
         unsafe {
-            zstd_sys::ZSTD_insertBlock(self.0, ptr_void(block), block.len())
+            zstd_sys::ZSTD_insertBlock(
+                self.0.as_ptr(),
+                ptr_void(block),
+                block.len(),
+            )
         }
     }
+
+    /// Creates a copy of this context.
+    ///
+    /// This only works before any data has been decompressed. An error will be
+    /// returned otherwise.
+    #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
+    pub fn try_clone(&self) -> Result<Self, ErrorCode> {
+        let context = NonNull::new(unsafe { zstd_sys::ZSTD_createDCtx() })
+            .ok_or(0usize)?;
+
+        unsafe { zstd_sys::ZSTD_copyDCtx(context.as_ptr(), self.0.as_ptr()) };
+
+        Ok(DCtx(context, self.1))
+    }
 }
 
 /// Prepares a new decompression context without dictionary.
@@ -897,7 +990,7 @@ pub fn create_dctx() -> DCtx<'static> {
 impl Drop for DCtx<'_> {
     fn drop(&mut self) {
         unsafe {
-            zstd_sys::ZSTD_freeDCtx(self.0);
+            zstd_sys::ZSTD_freeDCtx(self.0.as_ptr());
         }
     }
 }
@@ -936,50 +1029,60 @@ pub fn decompress_using_dict(
 }
 
 /// Compression dictionary.
-pub struct CDict<'a>(*mut zstd_sys::ZSTD_CDict, PhantomData<&'a ()>);
+pub struct CDict<'a>(NonNull<zstd_sys::ZSTD_CDict>, PhantomData<&'a ()>);
 
 impl CDict<'static> {
     pub fn create(
         dict_buffer: &[u8],
         compression_level: CompressionLevel,
     ) -> Self {
-        CDict(
-            unsafe {
+        Self::try_create(dict_buffer, compression_level)
+            .expect("zstd returned null pointer when creating dict")
+    }
+
+    pub fn try_create(
+        dict_buffer: &[u8],
+        compression_level: CompressionLevel,
+    ) -> Option<Self> {
+        Some(CDict(
+            NonNull::new(unsafe {
                 zstd_sys::ZSTD_createCDict(
                     ptr_void(dict_buffer),
                     dict_buffer.len(),
                     compression_level,
                 )
-            },
+            })?,
             PhantomData,
-        )
+        ))
     }
 }
 
 impl<'a> CDict<'a> {
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     pub fn create_by_reference(
         dict_buffer: &'a [u8],
         compression_level: CompressionLevel,
     ) -> Self {
         CDict(
-            unsafe {
+            NonNull::new(unsafe {
                 zstd_sys::ZSTD_createCDict_byReference(
                     ptr_void(dict_buffer),
                     dict_buffer.len(),
                     compression_level,
                 )
-            },
+            })
+            .expect("zstd returned null pointer"),
             PhantomData,
         )
     }
 
     pub fn sizeof(&self) -> usize {
-        unsafe { zstd_sys::ZSTD_sizeof_CDict(self.0) }
+        unsafe { zstd_sys::ZSTD_sizeof_CDict(self.0.as_ptr()) }
     }
 
     pub fn get_dict_id(&self) -> u32 {
-        unsafe { zstd_sys::ZSTD_getDictID_fromCDict(self.0) as u32 }
+        unsafe { zstd_sys::ZSTD_getDictID_fromCDict(self.0.as_ptr()) as u32 }
     }
 }
 
@@ -994,7 +1097,7 @@ pub fn create_cdict(
 impl<'a> Drop for CDict<'a> {
     fn drop(&mut self) {
         unsafe {
-            zstd_sys::ZSTD_freeCDict(self.0);
+            zstd_sys::ZSTD_freeCDict(self.0.as_ptr());
         }
     }
 }
@@ -1012,45 +1115,53 @@ pub fn compress_using_cdict(
     cctx.compress_using_cdict(dst, src, cdict)
 }
 
-pub struct DDict<'a>(*mut zstd_sys::ZSTD_DDict, PhantomData<&'a ()>);
+/// A digested decompression dictionary.
+pub struct DDict<'a>(NonNull<zstd_sys::ZSTD_DDict>, PhantomData<&'a ()>);
 
 impl DDict<'static> {
     pub fn create(dict_buffer: &[u8]) -> Self {
-        DDict(
-            unsafe {
+        Self::try_create(dict_buffer)
+            .expect("zstd returned null pointer when creating dict")
+    }
+
+    pub fn try_create(dict_buffer: &[u8]) -> Option<Self> {
+        Some(DDict(
+            NonNull::new(unsafe {
                 zstd_sys::ZSTD_createDDict(
                     ptr_void(dict_buffer),
                     dict_buffer.len(),
                 )
-            },
+            })?,
             PhantomData,
-        )
+        ))
     }
 }
 
 impl<'a> DDict<'a> {
     pub fn sizeof(&self) -> usize {
-        unsafe { zstd_sys::ZSTD_sizeof_DDict(self.0) }
+        unsafe { zstd_sys::ZSTD_sizeof_DDict(self.0.as_ptr()) }
     }
 
     /// Wraps the `ZSTD_createDDict_byReference()` function.
     ///
     /// The dictionary will keep referencing `dict_buffer`.
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     pub fn create_by_reference(dict_buffer: &'a [u8]) -> Self {
         DDict(
-            unsafe {
+            NonNull::new(unsafe {
                 zstd_sys::ZSTD_createDDict_byReference(
                     ptr_void(dict_buffer),
                     dict_buffer.len(),
                 )
-            },
+            })
+            .expect("zstd returned null pointer"),
             PhantomData,
         )
     }
 
     pub fn get_dict_id(&self) -> u32 {
-        unsafe { zstd_sys::ZSTD_getDictID_fromDDict(self.0) as u32 }
+        unsafe { zstd_sys::ZSTD_getDictID_fromDDict(self.0.as_ptr()) as u32 }
     }
 }
 
@@ -1064,7 +1175,7 @@ pub fn create_ddict(dict_buffer: &[u8]) -> DDict<'static> {
 impl<'a> Drop for DDict<'a> {
     fn drop(&mut self) {
         unsafe {
-            zstd_sys::ZSTD_freeDDict(self.0);
+            zstd_sys::ZSTD_freeDDict(self.0.as_ptr());
         }
     }
 }
@@ -1091,7 +1202,7 @@ pub type CStream<'a> = CCtx<'a>;
 
 /// Allocates a new `CStream`.
 pub fn create_cstream<'a>() -> CStream<'a> {
-    CCtx(unsafe { zstd_sys::ZSTD_createCStream() }, PhantomData)
+    CCtx::create()
 }
 
 /// Prepares an existing `CStream` for compression at the given level.
@@ -1160,6 +1271,7 @@ pub unsafe trait WriteBuf {
 }
 
 #[cfg(feature = "std")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "std")))]
 unsafe impl WriteBuf for std::vec::Vec<u8> {
     fn as_slice(&self) -> &[u8] {
         &self[..]
@@ -1176,6 +1288,7 @@ unsafe impl WriteBuf for std::vec::Vec<u8> {
 }
 
 #[cfg(feature = "arrays")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "arrays")))]
 unsafe impl<const N: usize> WriteBuf for [u8; N] {
     fn as_slice(&self) -> &[u8] {
         self
@@ -1513,6 +1626,7 @@ pub fn get_frame_content_size(src: &[u8]) -> u64 {
 ///
 /// `src` should be exactly a sequence of ZSTD frames.
 #[cfg(feature = "experimental")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
 pub fn find_decompressed_size(src: &[u8]) -> u64 {
     unsafe { zstd_sys::ZSTD_findDecompressedSize(ptr_void(src), src.len()) }
 }
@@ -1551,6 +1665,7 @@ pub fn sizeof_ddict(ddict: &DDict<'_>) -> usize {
 ///
 /// The dictionary will keep referencing `dict_buffer`.
 #[cfg(feature = "experimental")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
 pub fn create_cdict_by_reference<'a>(
     dict_buffer: &'a [u8],
     compression_level: CompressionLevel,
@@ -1560,6 +1675,7 @@ pub fn create_cdict_by_reference<'a>(
 
 /// Wraps the `ZSTD_isFrame()` function.
 #[cfg(feature = "experimental")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
 pub fn is_frame(buffer: &[u8]) -> u32 {
     unsafe { zstd_sys::ZSTD_isFrame(ptr_void(buffer), buffer.len()) as u32 }
 }
@@ -1568,6 +1684,7 @@ pub fn is_frame(buffer: &[u8]) -> u32 {
 ///
 /// The dictionary will keep referencing `dict_buffer`.
 #[cfg(feature = "experimental")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
 pub fn create_ddict_by_reference(dict_buffer: &[u8]) -> DDict {
     DDict::create_by_reference(dict_buffer)
 }
@@ -1592,7 +1709,8 @@ pub fn get_dict_id_from_frame(src: &[u8]) -> u32 {
 }
 
 /// Wraps the `ZSTD_initCStream_srcSize()` function.
-#[cfg(feacctxture = "experimental")]
+#[cfg(feature = "experimental")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
 #[deprecated]
 #[allow(deprecated)]
 pub fn init_cstream_src_size(
@@ -1605,6 +1723,7 @@ pub fn init_cstream_src_size(
 
 /// Wraps the `ZSTD_initCStream_usingDict()` function.
 #[cfg(feature = "experimental")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
 #[deprecated]
 #[allow(deprecated)]
 pub fn init_cstream_using_dict(
@@ -1617,6 +1736,7 @@ pub fn init_cstream_using_dict(
 
 /// Wraps the `ZSTD_initCStream_usingCDict()` function.
 #[cfg(feature = "experimental")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
 #[deprecated]
 #[allow(deprecated)]
 pub fn init_cstream_using_cdict<'a, 'b>(
@@ -1694,11 +1814,12 @@ pub fn cctx_reset(cctx: &mut CCtx<'_>, reset: ResetDirective) -> SafeResult {
 
 /// Wraps the `ZSTD_DCtx_reset()` function.
 pub fn dctx_reset(dctx: &mut DCtx<'_>, reset: ResetDirective) -> SafeResult {
-    parse_code(unsafe { zstd_sys::ZSTD_DCtx_reset(dctx.0, reset) })
+    parse_code(unsafe { zstd_sys::ZSTD_DCtx_reset(dctx.0.as_ptr(), reset) })
 }
 
 /// Wraps the `ZSTD_resetCStream()` function.
 #[cfg(feature = "experimental")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
 #[deprecated]
 #[allow(deprecated)]
 pub fn reset_cstream(zcs: &mut CStream, pledged_src_size: u64) -> SafeResult {
@@ -1707,6 +1828,7 @@ pub fn reset_cstream(zcs: &mut CStream, pledged_src_size: u64) -> SafeResult {
 
 /// Wraps the `ZSTD_initDStream_usingDict()` function.
 #[cfg(feature = "experimental")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
 #[deprecated]
 #[allow(deprecated)]
 pub fn init_dstream_using_dict(zds: &mut DStream, dict: &[u8]) -> SafeResult {
@@ -1715,6 +1837,7 @@ pub fn init_dstream_using_dict(zds: &mut DStream, dict: &[u8]) -> SafeResult {
 
 /// Wraps the `ZSTD_initDStream_usingDDict()` function.
 #[cfg(feature = "experimental")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
 #[deprecated]
 #[allow(deprecated)]
 pub fn init_dstream_using_ddict<'a, 'b>(
@@ -1729,11 +1852,13 @@ where
 
 /// Wraps the `ZSTD_resetDStream()` function.
 #[cfg(feature = "experimental")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
 pub fn reset_dstream(zds: &mut DStream) -> SafeResult {
     zds.reset()
 }
 
 #[cfg(feature = "experimental")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
 #[derive(Copy, Clone, Debug, PartialEq, Eq)]
 #[repr(u32)]
 pub enum FrameFormat {
@@ -1745,6 +1870,7 @@ pub enum FrameFormat {
 }
 
 #[cfg(feature = "experimental")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
 #[derive(Copy, Clone, Debug, PartialEq, Eq)]
 #[repr(u32)]
 pub enum DictAttachPref {
@@ -1756,6 +1882,7 @@ pub enum DictAttachPref {
 }
 
 #[cfg(feature = "experimental")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
 #[derive(Copy, Clone, Debug, PartialEq, Eq)]
 #[repr(u32)]
 pub enum ParamSwitch {
@@ -1768,48 +1895,63 @@ pub enum ParamSwitch {
 #[derive(Copy, Clone, Debug, PartialEq, Eq)]
 pub enum CParameter {
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     RSyncable(bool),
 
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     Format(FrameFormat),
 
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     ForceMaxWindow(bool),
 
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     ForceAttachDict(DictAttachPref),
 
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     LiteralCompressionMode(ParamSwitch),
 
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     TargetCBlockSize(u32),
 
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     SrcSizeHint(u32),
 
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     EnableDedicatedDictSearch(bool),
 
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     StableInBuffer(bool),
 
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     StableOutBuffer(bool),
 
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     BlockDelimiters(bool),
 
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     ValidateSequences(bool),
 
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     UseBlockSplitter(ParamSwitch),
 
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     UseRowMatchFinder(ParamSwitch),
 
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     DeterministicRefPrefix(bool),
 
     CompressionLevel(CompressionLevel),
@@ -1858,15 +2000,19 @@ pub enum DParameter {
 
     /// See `FrameFormat`.
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     Format(FrameFormat),
 
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     StableOutBuffer(bool),
 
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     ForceIgnoreChecksum(bool),
 
     #[cfg(feature = "experimental")]
+    #[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
     RefMultipleDDicts(bool),
 }
 
@@ -1929,12 +2075,14 @@ pub fn get_dict_id(dict_buffer: &[u8]) -> Option<u32> {
 
 /// Wraps the `ZSTD_getBlockSize()` function.
 #[cfg(feature = "experimental")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
 pub fn get_block_size(cctx: &CCtx) -> usize {
-    unsafe { zstd_sys::ZSTD_getBlockSize(cctx.0) }
+    unsafe { zstd_sys::ZSTD_getBlockSize(cctx.0.as_ptr()) }
 }
 
 /// Wraps the `ZSTD_compressBlock()` function.
 #[cfg(feature = "experimental")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
 pub fn compress_block(
     cctx: &mut CCtx,
     dst: &mut [u8],
@@ -1945,6 +2093,7 @@ pub fn compress_block(
 
 /// Wraps the `ZSTD_decompressBlock()` function.
 #[cfg(feature = "experimental")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
 pub fn decompress_block(
     dctx: &mut DCtx,
     dst: &mut [u8],
@@ -1955,12 +2104,14 @@ pub fn decompress_block(
 
 /// Wraps the `ZSTD_insertBlock()` function.
 #[cfg(feature = "experimental")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
 pub fn insert_block(dctx: &mut DCtx, block: &[u8]) -> usize {
     dctx.insert_block(block)
 }
 
 /// Wraps the `ZSTD_decompressBound` function
 #[cfg(feature = "experimental")]
+#[cfg_attr(feature = "doc-cfg", doc(cfg(feature = "experimental")))]
 pub fn decompress_bound(data: &[u8]) -> Result<u64, ErrorCode> {
     let bound =
         unsafe { zstd_sys::ZSTD_decompressBound(ptr_void(data), data.len()) };
